/************************************************************************
*
* disloader - Loads objects from asm980 for disassembly.
*
* Changes:
*   08/02/11   DGP   Original. Hacked from dis990
*   08/30/16   DGP   Label generation fixes.
*
************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#include "disdef.h"

extern FILE *lstfd;
extern int pgmlen;
extern int entry;
extern int modcount;
extern int symbolcount;
extern int cardmode;
extern char idtbuf[IDTSIZE+2];

extern uint16 memory[MEMSIZE];
extern Memory memctl[MEMSIZE];
extern Module modules[MAXMODULES];
extern SymNode *symbols[MAXSYMBOLS];

static int order;

/************************************************************************
* getnumfield - Gets the next numeric field.
************************************************************************/

static uint16
getnumfield (char *op, int binarymode)
{
   uint16 wdata;
   char item[80];

   if (binarymode)
   {
      wdata = ((*op & 0xFF) << 8) | (*(op+1) & 0xFF);
   }
   else
   {
      int wd;

      strncpy (item, op, 4);
      item[4] = '\0';
      sscanf (item, "%4X", &wd);
      wdata = wd & 0xFFFF;
   }
   return (wdata);
}

/************************************************************************
* getctl - Gets the Memory control address.
************************************************************************/

static Memory *
getctl (int mask, int address)
{
   return (&memctl[address]);
}

/************************************************************************
* getmem - Gets the contents of Memory.
************************************************************************/

static int
getmem (int mask, int address)
{
   return (GETMEM (address));
}

/************************************************************************
* putmem - Puts the value into Memory.
************************************************************************/

static void
putmem (int mask, int address, int val)
{
   PUTMEM (address, val);
}

/************************************************************************
* processdef - Process DEF.
************************************************************************/

static void
processdef (char *item, char *module, int relo, int flags, uint16 wdata)
{
   SymNode *s;
   char *bp;

   bp = item;
   for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
#ifdef DEBUGLOADER
   fprintf (stderr, ", %c%s = %s",
	    relo ? 'R' : 'A', "DEF", item);
#endif
   if ((s = symlookup (item, module, FALSE)) == NULL)
   {
      s = symlookup (item, module, TRUE);
      s->flags = relo ? RELOCATABLE : 0;
      s->flags |= GLOBAL | flags;
      s->value = wdata;
#ifdef DEBUGLOADER1
      fprintf (stderr, "\n      flags = >%04X, value = >%04X",
	       s->flags, s->value);
#endif
   }
}

/************************************************************************
* processref - Process REF.
************************************************************************/

static void
processref (char *item, char *module, int relo, int flags,
	    char otag, uint16 wdata)
{
   SymNode *s;
   char *bp;

   bp = item;
   for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
#ifdef DEBUGLOADER
   fprintf (stderr, ", %c%s = %s",
	    relo ? 'R' : 'A', "REF", item);
#endif

   if ((s = symlookup (item, module, FALSE)) == NULL)
   {
      s = symlookup (item, module, TRUE);
      s->flags = relo ? RELOCATABLE : 0;
      s->flags |= EXTERNAL | UNDEF | flags;
      s->value = wdata;
      s->order = order++;
#ifdef DEBUGLOADER
      fprintf (stderr, ", order = >%04X", s->order);
#endif
   }
}

/************************************************************************
* processrefchain - Process REF chain.
************************************************************************/

static void
processrefchain (int mask)
{
   int i;

#ifdef DEBUGLOADER
   fprintf (stderr, "processrefchain: mode = %s, mask = >%04X\n",
	    "P/R", mask);
#endif

   if (cardmode)
      return;

   for (i = 0; i < symbolcount; i++)
   {
      SymNode *s;

      s = symbols[i];
      if ((s->flags & mask) == mask)
      {
	 Memory *ctl;
	 int refaddr;

	 refaddr = s->value;
	 ctl = getctl (mask, refaddr);
#ifdef DEBUGLOADER
         fprintf (stderr, "   sym = %-6.6s, refaddr = >%04X\n",
		  s->symbol, refaddr);
#endif
	 while (refaddr)
	 {
	    int k;

	    k = getmem (mask, refaddr);
	    if (k == 0)
	    {
	       ctl->tag = RELDATA_TAG;
	    }
	    ctl->external = TRUE;
	    ctl->relocatable = TRUE;
	    putmem (mask, refaddr, s->value);
	    refaddr = k;
#ifdef DEBUGLOADER
	    fprintf (stderr, "                 refaddr = >%04X\n", refaddr);
#endif
	 }
      }
   }
}

/************************************************************************
* instlabels - Process and add labels for instructions.
************************************************************************/

static int
instlabels (uint16 inst, int *pc, int length, int mask, int refaddr,
	    char *module)
{
   SymNode *s;
   Memory *ctl;
   int type;
   int lpc;
   uint16 ixb;
   int16 wpc;
   int8 disp;

   lpc = *pc;
   type = decodeinst (inst);
   switch (type)
   {
   case TYPE_1:
      disp = inst & 0x00FF;
      ixb = (inst & 0x0700) >> 8;
#ifdef DEBUGLOADER
      fprintf (stderr,
	       "   TYPE_1 inst: pc = >%04X, ixb = %d, disp = >%02X\n",
	       lpc, ixb, (uint8)disp & 0xFF);
#endif
      if ((uint8)disp != 0xFF)
      {
	 wpc =  (int16)lpc + 1 + (int8)disp;
	 switch (ixb)
	 {
	 case 4:
	 case 0:
	 case 2:
	 case 6:
	    if (disp == 0)
	       wpc = getmem (mask, wpc);
#ifdef DEBUGLOADER
	    fprintf (stderr, "     wpc = >%04X, length = >%04X\n", wpc, length);
#endif
	    if ((wpc >= 0) && (wpc <= length) &&
		((s = vallookup (wpc, FALSE, FALSE)) == NULL))
	    {
#ifdef DEBUGLOADER
	       fprintf (stderr, "      label = %s%03d, pc = >%04X",
			module, refaddr+1, lpc);
#endif
	       if ((s = symadd (module, ++refaddr)) != NULL)
	       {
#ifdef DEBUGLOADER
		  fprintf (stderr, ", value = >%04X", wpc);
#endif
		  s->value = wpc;
	       }
#ifdef DEBUGLOADER
	       fprintf (stderr, "\n");
#endif
	    }
	    break;
	 default: ;
	 }
      }
      if ((disp == 0) && ((ixb == 0) || (ixb == 4) || (ixb == 6)))
	 lpc++;
      break;

   case TYPE_6:
      ixb = ((inst & 0x00F0) >> 4);
      if ((ixb == 0) || (ixb == 1) || (ixb == 4) || (ixb == 5))
         break;

   case TYPE_5:
   case TYPE_7:
      lpc++;
      ctl = getctl (mask, lpc);
      if (ctl->tag == RELDATA_TAG)
      {
	 inst = getmem (mask, lpc);
	 if ((s = vallookup (inst, TRUE, FALSE)) == NULL)
	 {
	    if ((s = symadd (module, ++refaddr)) != NULL)
	    {
	       s->value = inst;
	    }
	 }
      }

   default: ;
   }

   *pc = lpc;
   return (refaddr);
}

/************************************************************************
* processlabels - Process and add labels for recognizable targets.
************************************************************************/

static int
processlabels (int mask, int length, int refaddr, char *module)
{
   int pc;

#ifdef DEBUGLOADER
   fprintf (stderr, "processlabels: mode =  %s, mask = >%04X, length = >%X\n",
	    "P/R", mask, length);
#endif

   for (pc = 0; pc < length; )
   {
      SymNode *s;
      Memory *ctl;
      uint16 wdata;
      char otag;

      ctl = getctl (mask, pc);
      otag = ctl->tag;

      if (otag == RELDATA_TAG)
      {
	 if (!ctl->external && !ctl->global)
	 {
	    wdata = getmem (mask, pc);
	    if ((s = vallookup (wdata, FALSE, FALSE)) == NULL)
	    {
#ifdef DEBUGLOADER
		  fprintf (stderr, "   RLABEL = %s%03d, pc = >%04X",
			   module, refaddr+1, pc);
#endif
		  if ((s = symadd (module, ++refaddr)) != NULL)
		  {
		     s->value = getmem (mask, pc);
#ifdef DEBUGLOADER
		     fprintf (stderr, ", value = >%04X", s->value);
#endif
		  }
#ifdef DEBUGLOADER
		  fprintf (stderr, "\n");
#endif
	    }
	 }
         pc++;
      }

      else if (otag == ABSDATA_TAG)
      {
	 if ((s = vallookup (pc, FALSE, FALSE)) != NULL)
	 {
	    if ((s->flags & mask) == mask)
	    {
#ifdef DEBUGLOADER
	       fprintf (stderr, "   entry  = %-6.6s, pc = >%04X\n",
			s->symbol, pc);
#endif
	       strcpy (module, s->symbol);
	       if (strlen (module) > 3) module[3] = '\0';
	    }
	 }
	 wdata = getmem (mask, pc);
	 if (wdata)
	    refaddr = instlabels (wdata, &pc, length, mask, refaddr, module); 
         pc++;
      }

      else if (otag == RELORG_TAG)
      {
         pc = getmem (mask, pc); 
      }

      else
      {
	 pc++;
      }
   }
   return (refaddr);
}

/***********************************************************************
* loadcardfile - Load a card mode object file.
***********************************************************************/

static int
loadcardfile (FILE *fd, int loadpt, char *file)
{
   SymNode *s;
   Memory *ctl;
   int idtrec = TRUE;
   int curraddr;
   int reloprog;
   int reclen;
   int refcnt, defcnt;
   uint16 reloword;
   uint16 pseglen;
   uint8 inbuf[82];
   char module[MAXSYMLEN+2];

#ifdef DEBUGLOADER
   fprintf (stderr, "\nloadcardfile: loadpt = %04X, file = %s\n",
	   loadpt, file);
#endif

   curraddr = loadpt;
   defcnt = refcnt = 0;
   pseglen = 0;
   while (fgets ((char *)inbuf, 82, fd) != NULL)
   {
      uint8 *op = (uint8 *)inbuf;
      uint16 wdata;
      char item[80];

      reclen = strlen ((char *)inbuf);

      /*
      ** The first "card" is the IDT record.
      */

      if (idtrec)
      {
	 int i;
	 char *tp;

         idtrec = FALSE;

	 /* Sanity check */

	 tp = op + IDTSIZE;
	 if (*tp != '0' && *tp != '1')
	    goto BAD_OBJECT;
	 tp++;
	 for (i = 0; i < 8; i++)
	    if (!isxdigit(*tp++))
	       goto BAD_OBJECT;
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    defcnt = (defcnt * 10) + (*tp - '0');
	    tp++;
	 }
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    refcnt = (refcnt * 10) + (*tp - '0');
	    tp++;
	 }

	 strncpy (item, (char *)op, IDTSIZE);
	 op += IDTSIZE;
	 item[IDTSIZE] = '\0';
#ifdef DEBUGLOADER
	 fprintf (stderr, "   IDT = %s", item);
#endif
	 tp = item;
	 for (; *tp; tp++) if (isspace(*tp)) *tp = '\0';
	 strcpy (module, item);
	 reloprog = *op++ == '1' ? TRUE : FALSE;
	 wdata = getnumfield (op, FALSE);
	 pseglen = wdata;
#ifdef DEBUGLOADER
	 fprintf (stderr, ", length = %04X", pseglen);
#endif
	 strcpy (modules[modcount].objfile, file);
	 strcpy (modules[modcount].name, item);
	 modules[modcount].length = wdata;
	 modules[modcount].origin = loadpt;
	 modules[modcount].creator[0] = '\0';
	 modules[modcount].date[0] = '\0';
	 modules[modcount].time[0] = '\0';
#ifdef DEBUGLOADER
	 fprintf (stderr, ", module = %s, modcount = %d", module, modcount);
#endif
	 if (idtbuf[0] == '\0')
	    strcpy (idtbuf, item);
      }

      /*
      ** Process DEF records
      */

      else if (defcnt)
      {
#ifdef DEBUGLOADER
	 fprintf (stderr, "   DEF: \n");
#endif
	 while (defcnt)
	 {
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    op += MAXSYMLEN;
	    wdata = getnumfield (op, FALSE);
	    op += WORDLEN;
	    if (reloprog)
	       wdata += loadpt;
#ifdef DEBUGLOADER
	    fprintf (stderr, "      sym = %s, val = %04X", item, wdata);
#endif
	    processdef (item, module, reloprog, 0, wdata);
#ifdef DEBUGLOADER
	    fprintf (stderr, "\n");
#endif
	    defcnt--;
	    if ((op-inbuf) >= 70)
	       break;
	 }
      }

      /*
      ** Process REF records
      */

      else if (refcnt)
      {
#ifdef DEBUGLOADER
	 fprintf (stderr, "   REF: \n");
#endif
	 while (refcnt)
	 {
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    op += MAXSYMLEN;
#ifdef DEBUGLOADER
	    fprintf (stderr, "      sym = %s", item);
#endif
	    processref (item, module, reloprog, 0, 0, 0);
#ifdef DEBUGLOADER
	    fprintf (stderr, "\n");
#endif
	    refcnt--;
	    if ((op-inbuf) >= 72)
	       break;
	 }
      }

      /*
      ** Process loadblocks
      */

      else
      {
	 int i, j;

#ifdef DEBUGLOADER
	 fprintf (stderr, "   LOAD BLOCK: loadpt = %04X, curraddr = %04X",
		 loadpt, curraddr);
#endif

	 if (!strncmp(op, "/*", 2))
	 {
	    if (reclen > 60)
	    {
	       strncpy (modules[modcount].date, (char *)&inbuf[TIMEOFFSET], 8);
	       modules[modcount].date[9] = '\0';
	       strncpy (modules[modcount].time, (char *)&inbuf[DATEOFFSET], 8);
	       modules[modcount].time[9] = '\0';
	       strncpy (modules[modcount].creator,
			(char *)&inbuf[CREATOROFFSET], 8);
	       modules[modcount].creator[9] = '\0';
	    }
	    else
	    {
	       strncpy (modules[modcount].date, "        ", 8);
	       strncpy (modules[modcount].time, "        ", 8);
	       strncpy (modules[modcount].creator, "        ", 8);
	    }
	    modcount++;
	    if (curraddr < pseglen)
	    {
	       wdata = pseglen;
	       if (reloprog)
	          wdata += loadpt;
#ifdef DEBUGLOADER
	       fprintf (stderr, ", ORG = %04X", wdata);
#endif
	       memctl[curraddr].tag = reloprog ? RELORG_TAG : ABSORG_TAG;
	       PUTMEM (curraddr, wdata);
	    }
	    curraddr = pseglen + loadpt;
	    loadpt = curraddr;
	    idtrec = TRUE;
	    continue;
	 }

	 if (!strncmp(op, "&&&&", 4))
	 {
	    op += WORDLEN;
	    wdata = getnumfield (op, FALSE);
#ifdef DEBUGLOADER
	    fprintf (stderr, ", ENTRY = %04X\n", wdata);
#endif
	    if (reloprog)
	       entry = (wdata + loadpt) & 0xFFFF;
	    else
	       entry = wdata & 0xFFFF;
	    continue;
	 }

         wdata = getnumfield (op, FALSE);
	 if (reloprog)
	    wdata += loadpt;
	 op += WORDLEN;
	 if (wdata > curraddr)
	 {
#ifdef DEBUGLOADER
	    fprintf (stderr, ", ORG = %04X", wdata);
#endif
	    memctl[curraddr].tag = reloprog ? RELORG_TAG : ABSORG_TAG;
	    PUTMEM (curraddr, wdata);
	 }
	 curraddr = wdata;

	 for (i = 0; i < 2; i++)
	 {
	    reloword = getnumfield (op, FALSE);
	    op += WORDLEN;
	    for (j = 0; j < 8; j++)
	    {
	       int relo;

	       if (*op == ' ')
	       {
	          i = j = 9;
		  continue;
	       }
	       wdata = getnumfield (op, FALSE);
	       op += WORDLEN;
	       relo = (reloword >> (14 - (j*2))) & 3;
	       switch (relo)
	       {
	       case 0: /* Absolute */
		  ctl = getctl (FALSE, curraddr);
		  putmem (FALSE, curraddr, wdata);
		  ctl->tag = ABSDATA_TAG;
		  ctl->relocatable = 0;
		  curraddr++;
		  break;

	       case 2: /* Relocatable */
	          wdata += loadpt;
		  ctl = getctl (FALSE, curraddr);
		  putmem (FALSE, curraddr, wdata);
		  ctl->tag = RELDATA_TAG;
		  ctl->relocatable = reloprog;
		  curraddr++;
		  break;

	       case 3: /* External */
		  ctl = getctl (FALSE, curraddr);
		  putmem (FALSE, curraddr, wdata);
		  ctl->tag = EXTVALUE_TAG;
		  ctl->relocatable = reloprog;
		  ctl->external = TRUE;
		  curraddr++;
	          break;

	       default:
	          fprintf (stderr, "Unsupported relocation %d\n", relo);
		  exit (ABORT);
	       }
	    }
	 }
      }
#ifdef DEBUGLOADER
      fprintf (stderr, "\n");
#endif
   }
#ifdef DEBUGLOADER
   fprintf (stderr, "\n");
#endif
   return (curraddr);

BAD_OBJECT:
   fprintf (stderr, "Bad object file\n");
   exit (ABORT);
}

/************************************************************************
* disloader - Object loader for dis980.
************************************************************************/

int
disloader (FILE *fd, int loadpt, char *file)
{
   SymNode *s;
   Memory *ctl;
   int i;
   int done;
   int status = 0;
   int binarymode = FALSE;
   int reclen;
   int curraddr;
   int origin;
   int relo;
   int refaddr;
   int wordlen = WORDTAGLEN - 1;
   uint16 wdata;
   char inbuf[82];
   char module[MAXSYMLEN+2];

   order = 0;
   i = fgetc (fd);
   ungetc (i, fd);
   if (i != IDT_TAG)
   {
      cardmode = TRUE;
      curraddr = loadcardfile (fd, loadpt, file);
      goto BACK_END;
   }

#ifdef DEBUGLOADER
   fprintf (stderr, "disloader: loadpt = %04X, file = %s\n",
	    loadpt, file);
#endif

   cardmode = FALSE;
   origin = curraddr = loadpt;

   done = FALSE;
   while (!done)
   {
      char *op = inbuf;
      int i;

      if (fgets (inbuf, 82, fd) == NULL)
      {
	 done = TRUE;
	 break;
      }
      reclen = strlen (inbuf);

#ifdef DEBUGLOADER
      fprintf (stderr, "New record, reclen = %d\n", reclen);
#endif
      if (*op == EOFSYM)
      {
	 if (reclen > 60)
	 {
	    strncpy (modules[modcount].date, &inbuf[TIMEOFFSET], 8);
	    modules[modcount].date[9] = '\0';
	    strncpy (modules[modcount].time, &inbuf[DATEOFFSET], 8);
	    modules[modcount].time[9] = '\0';
	    strncpy (modules[modcount].creator, &inbuf[CREATOROFFSET], 8);
	    modules[modcount].creator[9] = '\0';
	 }
	 modcount++;
	 loadpt = curraddr;
	 continue;
      }

      for (i = 0; i < reclen; i++)
      {
	 char *bp;
	 char otag;
	 char item[80];

	 relo = FALSE;
	 otag = *op++;
	 if (otag == EOR_TAG)
	    wdata = 0;
	 else
	    wdata = getnumfield (op, binarymode);

#ifdef DEBUGLOADER
	 fprintf (stderr,
	       "   otag = %c, loadpt = >%04X, curraddr = >%04X, data = >%04X",
		  isprint(otag) ? otag : '0', loadpt, curraddr, wdata);
#endif

	 refaddr = 0;

         switch (otag)
	 {
	 case IDT_TAG:
	    op += wordlen;
	    strncpy (item, op, IDTSIZE);
	    item[IDTSIZE] = '\0';
#ifdef DEBUGLOADER
	    fprintf (stderr, ", IDT = %s", item);
#endif
	    bp = item;
	    for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
	    strcpy (module, item);
	    strcpy (modules[modcount].objfile, file);
	    strcpy (modules[modcount].name, item);
	    modules[modcount].length = wdata;
	    modules[modcount].origin = loadpt;
	    modules[modcount].creator[0] = '\0';
	    modules[modcount].date[0] = '\0';
	    modules[modcount].time[0] = '\0';
	    if (idtbuf[0] == '\0')
	    {
	       strcpy (idtbuf, item);
	    }
	    op += IDTSIZE;
	    break;

         case RELORG_TAG:
	    wdata += loadpt;
         case ABSORG_TAG:
	    PUTMEM (curraddr, wdata);
	    memctl[curraddr].tag = otag;
	    curraddr = wdata & 0xFFFF;
	    op += wordlen;
	    break;

	 case RELDATA_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSDATA_TAG:
	    ctl = getctl (FALSE, curraddr);
	    putmem (FALSE, curraddr, wdata);
	    curraddr++;
	    ctl->tag = otag;
	    ctl->relocatable = relo;
	    op += wordlen;
	    break;

	 case RELENTRY_TAG:
	    entry = (wdata + loadpt) & 0xFFFF;
	    op += wordlen;
	    break;

	 case ABSENTRY_TAG:
	    entry = wdata & 0xFFFF;
	    op += wordlen;
	    break;

	 case RELEXTRN_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSEXTRN_TAG:
	    refaddr = wdata;
	    op += wordlen;
	    strncpy (item, op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    processref (item, module, relo, 0, otag, wdata);
	    op += MAXSYMLEN;
	    break;

	 case RELGLOBAL_TAG:
	    wdata += loadpt;
	    relo = TRUE;
	 case ABSGLOBAL_TAG:
	    op += wordlen;
	    strncpy (item, op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    processdef (item, module, relo, 0, wdata);
	    op += MAXSYMLEN;
	    break;

	 case EOR_TAG:
	    i = 81;
	    break;

	 default: ;
	    op += wordlen;
	 }
#ifdef DEBUGLOADER
	 fprintf (stderr, "\n");
#endif
      }
   }

BACK_END:
   /*
   ** Go through external chain and set value for symbol.
   */

   processrefchain (RELOCATABLE | EXTERNAL);

   /*
   ** Go through the code and generate relocatable symbols.
   */

   strcpy (module, idtbuf);
   if (strlen (module) > 3) module[3] = '\0';

   refaddr = processlabels (GLOBAL, curraddr, 0, module);

   if (entry >= 0)
   {
      if ((s = vallookup (entry, FALSE, FALSE)) == NULL)
      {
#ifdef DEBUGLOADER
	 fprintf (stderr, "   ELABEL = %s%03d, entry = >%04X",
		  module, refaddr+1, entry);
#endif
	 if ((s = symadd (module, ++refaddr)) != NULL)
	 {
	    s->value = entry;
	 }
      }
   }
      
#ifdef DEBUGLOADER
   fprintf (stderr, "\nSymbols:\n");
   for (i = 0; i < symbolcount; i++)
   {
      s = symbols[i];
      fprintf (stderr,
	       " symbol[%d]: '%-8.8s' '%-8.8s' flags = >%04X, value = >%04X\n",
	       i, s->symbol, s->module, s->flags, s->value);
   }
#endif

   pgmlen = curraddr;
   return (status);
}
