/***********************************************************************
*
* disops.c - Print instructions for the TI 980 computer.
*
* Changes:
*   08/02/11   DGP   Original. Hacked from dis990.
*   09/27/11   DGP   Added cpu models and Calder 980 support.
*   08/25/16   DGP   Fixed MOV modes decode.
*   08/30/16   DGP   Lots of fixes...
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include "disdef.h"

extern int pcreg;	/* The program PC */
extern int pgmlen;
extern int symbolcount;
extern int cardmode;
extern int model;

extern char idtbuf[IDTSIZE+2];

extern uint16 memory[MEMSIZE];
extern Memory memctl[MEMSIZE];
extern SymNode *symbols[MAXSYMBOLS];

static char outbuf[82];
static int extended;
static int extendaddr;

/*
** Opcode tables.
*/

static char *ops1op[] =
{
   "LDA",  "LDE",  "LDX",  "LDM",  "ADD",  "SUB",  "IOR",  "AND",
   "BIX",  "DMT",  "IMO",  "DIV",  "CPL",  "CPA",  "BRL",  "BRU",
   "STA",  "STE",  "STX",  "MPY",  "DST",  "DSB",  "DLD",  "DAD"
};

static char *ops2op[] =
{
   "RSU",  "RAD",  "RCO",  "",     "RIV",  "REO",  "RIN",  "",
   "RCA",  "ROR",  "RMO",  "",     "RCL",  "RAN",  "RDE",  "REX"
};

static char *ops3op[] =
{
   "ARA",  "ARD",  "LRA",  "LRD",  "ALA",  "ALD",  "LLA",  "LLD",
   "RTO",  "",     "RTZ",  "",     "LTO",  "",     "LTZ",  "",
   "CRA",  "CRE",  "CRX",  "CRM",  "NRM",  "",     "",     "", 
   "",     "CRS",  "CRL",  "CRB",  "CLD",  "",     "CRD",
};

static char *ops4op[] =
{
   "SZE",  "SSE",  "SOO",  "",     "SOD",  "",     "SMI",  "",
   "SNZ",  "SSN",  "SNO",  "",     "SEV",  "",     "SPL",  "",
   "SLT",  "",     "SEQ",  "",     "SGT",  "",     "SOV",  "",
   "SGE",  "",     "SNE",  "",     "SLE",  "",     "SNV",  "",
   "IDL",  "SOC",  "SNC",
};

static char *ops5op[] =
{
   "RDS",  "",     "WDS",  "",     "",     "",     "",     "",
   "LSB",  "LSR",  "LRF",  "",     "SSB",  "",     "SRF",
};

static char *ops6op[] =
{
   "TABZ", "TABO", "TMBZ", "TMBO", "SABZ", "SABO", "SMBZ", "SMBO",
};

static char *ops7op[] =
{
   "ATI",  "API"
};

static char *ops8op[] =
{
   "MVC",  "CLC"
};

static char *ops10op[] =
{
   "LAR", "SAR", "RFI", "RFD"
};

static char *ops11op[] =
{
   "SWTX", "SWT", "SWTS"
};

static char *movmodes[] =
{
   "*", "*+", "-*", "", "#", "#+", "-#", ""
};

extern char *regnames[16];

/*
** Function definitions and branch table
*/

static int proc1op  (uint16 inst);
static int proc2op  (uint16 inst);
static int proc3op  (uint16 inst);
static int proc4op  (uint16 inst);
static int proc5op  (uint16 inst);
static int proc6op  (uint16 inst);
static int proc7op  (uint16 inst);
static int proc8op  (uint16 inst);
static int proc9op  (uint16 inst);
static int proc10op  (uint16 inst);
static int proc11op  (uint16 inst);

typedef struct {
   int (*proc) (uint16 inst);
} Inst_Proc;

Inst_Proc inst_proc[MAX_INST_TYPES] = {
   proc1op,  proc2op,  proc3op,  proc4op,  proc5op,  proc6op,  proc7op,
   proc8op,  proc9op,  proc10op, proc11op
};

/***********************************************************************
* printvals - Print symbols whose value equal val.
***********************************************************************/

static void
printvals (int val, int inc)
{
   SymNode *s;
   int i;

   for (i = 0; i < symbolcount; i++)
   {
      s = symbols[i];
      if (!(s->flags & EXTERNAL) && (s->flags & RELOCATABLE) && 
	   (s->value == val))
      {
	 printf ("%s\tEQU\t$", s->symbol);
	 if (inc > 0)
	    printf ("+%d", inc);
	 else if (inc < 0)
	    printf ("%d", inc);
         printf ("\n");
      }
   }
}

/***********************************************************************
* decodeinst - Decode instructions.
***********************************************************************/

int
decodeinst (uint16 inst)
{
   uint16 op;

   if ((inst & 0xF800) < 0xC000)
   {
      op = (inst >> 11);
      switch (op)
      {
      case 0:  /* LDA */
      case 1:  /* LDE */
      case 2:  /* LDX */
      case 3:  /* LDM */
      case 4:  /* ADD */
      case 5:  /* SUB */
      case 6:  /* IOR */
      case 7:  /* AND */
      case 8:  /* BIX */
      case 9:  /* DMT */
      case 10: /* IMO */
      case 11: /* DIV */
      case 12: /* CPL */
      case 13: /* CPA */
      case 14: /* BRL */
      case 15: /* BRU */
      case 16: /* STA */
      case 17: /* STE */
      case 18: /* STX */
      case 19: /* MPY */
      case 20: /* DST */
      case 21: /* DSB */
      case 22: /* DLD */
      case 23: /* DAD */
	 return (TYPE_1);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xC800)
   {
      op = ((inst & 0x0F80) >> 7);
      switch (op)
      {
      case 0:  /* RSU */
      case 1:  /* RAD */
      case 2:  /* RCO */
      case 4:  /* RIV */
      case 5:  /* REO */
      case 6:  /* RIN */
      case 8:  /* RCA */
      case 9:  /* ROR */
      case 10: /* RMO */
      case 12: /* RCL */
      case 13: /* RAN */
      case 14: /* RDE */
      case 15: /* REX */
	 return (TYPE_2);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xCC00)
   {
      op = ((inst & 0x03F0) >> 5);
      switch (op)
      {
      case 0:  /* ARA */
      case 1:  /* ARD */
      case 2:  /* LRA */
      case 3:  /* LRD */
      case 4:  /* ALA */
      case 5:  /* ALD */
      case 6:  /* LLA */
      case 7:  /* LLD */
      case 8:  /* RTO */
      case 10: /* RTZ */
      case 12: /* LTO */
      case 14: /* LTZ */
      case 16: /* CRA */
      case 17: /* CRE */
      case 18: /* CRX */
      case 19: /* CRM */
      case 20: /* NRM */
      case 25: /* CRS */
      case 26: /* CRL */
      case 27: /* CRB */
      case 28: /* CLD */
      case 30: /* CRD */
	 return (TYPE_3);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xD000)
   {
      op =  ((inst & 0x3F0) >> 4);
      switch (op)
      {
      case 0:  /* SZE */
      case 1:  /* SSE */
      case 2:  /* SOO */
      case 4:  /* SOD */
      case 6:  /* SMI */
      case 8:  /* SNZ */
      case 9:  /* SSN */
      case 10: /* SNO */
      case 12: /* SEV */
      case 14: /* SPL */
      case 16: /* SLT */
      case 18: /* SEQ */
      case 20: /* SGT */
      case 22: /* SOV */
      case 24: /* SGE */
      case 26: /* SNE */
      case 28: /* SLE */
      case 30: /* SNV */
      case 32: /* IDL */
      case 54: /* SOC */
      case 62: /* SNC */
	 return (TYPE_4);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xD900)
   {
      op = ((inst & 0x00F0) >> 4);
      if ((inst & 0xD9A0) == 0xD800) /* RDS */
	 return (TYPE_5);
      else if ((inst & 0xD9A0) == 0xD820) /* WDS */
	 return (TYPE_5);
      else switch (op)
      {
      case 8:  /* LSB */
      case 12: /* SSB */
	 return (TYPE_5);

      case 9:  /* LSR */
      case 10: /* LRF */
      case 14: /* SRF */
	 if (model >= CPU_TI980A)
	    return (TYPE_5);

      default: ;
      }
   }
   else if ((inst & 0xFF00) == 0xD900)
   {
      if ((inst & 0xFF00) == 0xD900) /* ATI */
	 return (TYPE_7);
   }
   else if ((inst & 0xFF00) == 0xDB00)
   {
      if (model >= CPU_TI980A)
      {
	 op = ((inst & 0x00F0) >> 4);
	 switch (op)
	 {
	 case 0: /* TABZ */
	 case 1: /* TABO */
	 case 4: /* SABZ */
	 case 5: /* SABO */
	 case 2: /* TMBZ */
	 case 3: /* TMBO */
	 case 6: /* SMBZ */
	 case 7: /* SMBO */
	    return (TYPE_6);

	 default: ;
	 }
      }
   }
   else if ((inst & 0xFF00) == 0xDD00)
   {
      if (model >= CPU_TI980A)
      {
	 if ((inst & 0xFF00) == 0xDD00) /* API */
	    return (TYPE_7);
      }
   }
   else if ((inst & 0xFF00) == 0xDF00)
   {
      if (model >= CPU_TI980A)
      {
	 if ((inst & 0xFF80) == 0xDF00) /* MVC */
	    return (TYPE_8);
	 else if ((inst & 0xFF80) == 0xDF80) /* CLC */
	    return (TYPE_8);
      }
   }
   else if ((inst & 0xF000) == 0xE000)
   {
      if (model == CPU_CDD980)
	 return (TYPE_9);
   }
   else if ((inst & 0xFF00) == 0xF800)
   {
      if (model == CPU_CDD980)
      {
	 if (((inst & 0x0088) == 0x0088) || /* LAR */
	     ((inst & 0x0088) == 0x0080) || /* SAR */
	     ((inst & 0x0088) == 0x0008) || /* RFI */
	     ((inst & 0x0088) == 0x0000))   /* RFD */
	    return (TYPE_10);
      }
   }
   else if ((inst & 0xF000) == 0xF000)
   {
      if (model == CPU_CDD980)
      {
	 if (((inst & 0x0F00) == 0x0000) || /* SWTX */
	     ((inst & 0x0F00) == 0x0100) || /* SWT */
	     ((inst & 0x0F00) == 0x0200))   /* SWTS */
	    return (TYPE_11);
      }
   }
   return (TYPE_ILL);
}

/***********************************************************************
* getctl - Get mem control address.
***********************************************************************/

static Memory *
getctl (void)
{
   return (&memctl[pcreg]);
}

/***********************************************************************
* getinst - Get an Instruction.
***********************************************************************/

static uint16
getinst (void)
{
   uint16 inst;

   inst = GETMEM (pcreg);
   pcreg++;

   return (inst);
}

/***********************************************************************
* procmemop - Process memory operand.
***********************************************************************/

static void
procmemop (char *operand, Memory *operctl, uint16 oper)
{
   SymNode *sym;

   if (operctl->tag == RELDATA_TAG)
   {
      if (operctl->external)
	 sym = extlookup (oper, cardmode);
      else 
	 sym = vallookup (oper, TRUE, FALSE);

      if (sym)
	 strcpy (operand, sym->symbol);
      else
	 sprintf (operand, ">%X", oper);
   }
   else
   {
      sprintf (operand, ">%X", oper);
   }
}

/***********************************************************************
* procdispop - Process displacement operand.
***********************************************************************/

static void
procdispop (char *operand, int16 wpc, uint8 disp, char *ixb, char *ind)
{
   SymNode *sym;

   sym = vallookup (wpc, TRUE, FALSE);
   if (sym)
      sprintf (operand, "%s%s%s", ind, sym->symbol, ixb);
   else
      sprintf (operand, "%s>%X%s", ind, disp, ixb);
}

/***********************************************************************
* proc1op - Process type 1 operations.
***********************************************************************/

static int
proc1op (uint16 inst)
{
   SymNode *sym;
   Memory *operctl;
   uint16 ixb;
   uint16 op;
   uint16 oper;
   int16 wpc;
   uint8 disp;
   char temp[32];
   char operand[32];

   ixb = (inst & 0x0700) >> 8;
   disp = (inst & 0x00FF);
   op = (inst >> 11);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   extended = FALSE;
   if ((disp == 0) && ((ixb == 0) || (ixb == 4) || (ixb == 6)))
   {
      extended = TRUE;
      extendaddr = pcreg;
   }

   switch (op)
   {
   case 0: /* LDA */
   case 1: /* LDE */
   case 2: /* LDX */
   case 3: /* LDM */
   case 4: /* ADD */
   case 5: /* SUB */
   case 6: /* IOR */
   case 7: /* AND */
   case 8: /* BIX */
   case 9: /* DMT */
   case 10: /* IMO */
   case 11: /* DIV */
   case 12: /* CPL */
   case 13: /* CPA */
   case 14: /* BRL */
   case 15: /* BRU */
   case 16: /* STA */
   case 17: /* STE */
   case 18: /* STX */
   case 19: /* MPY */
   case 20: /* DST */
   case 21: /* DSB */
   case 22: /* DLD */
   case 23: /* DAD */
      break;

   default:
      return (-1);
   }

   if (extended)
      sprintf (temp, "@%s", ops1op[op]);
   else
      strcpy (temp, ops1op[op]);

   switch (ixb)
   {
   case 0:	/* PC relative */
      if (disp == 0)
      {
	 operctl = getctl();
	 if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
	 {
	    extended = FALSE;
	    return (-1);
	 }
	 oper = getinst();
	 operand[0] = '=';
	 procmemop (&operand[1], operctl, oper);
      }
      else
      {
	 if ((int16)pcreg + (int8)disp < 0)
	 {
	    extended = FALSE;
	    return (-1);
	 }
	 wpc = (int16)pcreg + (int8)disp;
	 if (wpc >= 0 && wpc <= pgmlen)
	 {
	    sym = vallookup (wpc, TRUE, FALSE);
	    if (sym)
	       strcpy (operand, sym->symbol);
	    else if (disp < 128)
	       sprintf (operand, "$+%d", disp+1);
	    else 
	       sprintf (operand, "$-%d", (int8)disp+1);
	 }
	 else
	 {
	    extended = FALSE;
	    return (-1);
	 }
      }
      break;

   case 1:	/* Base register relative */
      sprintf (operand, ">%X,1", disp);
      break;

   case 2:	/* Indexed PC relative */
      if ((int16)pcreg + (int8)disp < 0)
      {
	 extended = FALSE;
	 return (-1);
      }
      wpc = (int16)pcreg + (int8)disp;
      if (wpc >= 0 && wpc <= pgmlen)
      {
	 procdispop (operand, wpc, disp, ",2", "");
      }
      else
      {
	 extended = FALSE;
	 return (-1);
      }
      break;

   case 3:	/* Indexed base register relative */
      sprintf (operand, ">%X,3", disp);
      break;

   case 4:	/* Indirect PC relative */
      if (disp == 0)
      {
	 operctl = getctl();
	 if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
	 {
	    extended = FALSE;
	    return (-1);
	 }
	 oper = getinst();
	 procmemop (operand, operctl, oper);
      }
      else
      {
	 if ((int16)pcreg + (int8)disp < 0)
	 {
	    extended = FALSE;
	    return (-1);
	 }
	 wpc = (int16)pcreg + (int8)disp;
	 if (wpc >= 0 && wpc <= pgmlen)
	 {
	    procdispop (operand, wpc, disp, "", "*");
	 }
	 else
	 {
	    extended = FALSE;
	    return (-1);
	 }
      }
      break;

   case 5:	/* Indirect base register relative */
      sprintf (operand, "*>%X,1", disp);
      break;

   case 6:	/* Indirect, indexed, PC relative */
      if ((int16)pcreg + (int8)disp < 0)
      {
	 extended = FALSE;
	 return (-1);
      }
      wpc = (int16)pcreg + (int8)disp;
      if (wpc >= 0 && wpc <= pgmlen)
      {
	 procdispop (operand, wpc, disp, ",2", "*");
      }
      else
      {
	 extended = FALSE;
	 return (-1);
      }
      break;

   case 7:	/* Immediate */
      sprintf (operand, "=%d", (int8)disp);
   }

   sprintf (outbuf, "\t%s\t%s", temp, operand);

   return (0);
}

/***********************************************************************
* proc2op - Process type 2 operations.
***********************************************************************/

static int
proc2op (uint16 inst)
{
   uint16 op;
   uint16 sr, dr;

   dr = inst & 0x0007;
   sr = (inst >> 4) & 0x0007;
   op = ((inst & 0x0F80) >> 7);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* RSU */
   case 1: /* RAD */
   case 2: /* RCO */
   case 4: /* RIV */
   case 5: /* REO */
   case 6: /* RIN */
   case 8: /* RCA */
   case 9: /* ROR */
   case 10: /* RMO */
   case 12: /* RCL */
   case 13: /* RAN */
   case 14: /* RDE */
   case 15: /* REX */
      break;

   default:
      return (-1);
   }

   sprintf (outbuf, "\t%s\t%s,%s", ops2op[op], regnames[sr], regnames[dr]);
   return (0);
}

/***********************************************************************
* proc3op - Process type 3 operations.
***********************************************************************/

static int
proc3op (uint16 inst)
{
   uint16 count;
   uint16 op;

   count = inst & 0x001F;
   op = ((inst & 0x03F0) >> 5);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* ARA */
   case 1: /* ARD */
   case 2: /* LRA */
   case 3: /* LRD */
   case 4: /* ALA */
   case 5: /* ALD */
   case 6: /* LLA */
   case 7: /* LLD */
   case 8: /* RTO */
   case 10: /* RTZ */
   case 12: /* LTO */
   case 14: /* LTZ */
   case 16: /* CRA */
   case 17: /* CRE */
   case 18: /* CRX */
   case 19: /* CRM */
   case 25: /* CRS */
   case 26: /* CRL */
   case 27: /* CRB */
   case 28: /* CLD */
   case 30: /* CRD */
      break;

   case 20: /* NRM */
      sprintf (outbuf, "\t%s", ops3op[op]);
      return (0);

   default:
      return (-1);
   }

   sprintf (outbuf, "\t%s\t%d", ops3op[op], count);
   return (0);
}

/***********************************************************************
* proc4op - Process type 4 operations.
***********************************************************************/

static int
proc4op (uint16 inst)
{
   uint16 op;

   op =  ((inst & 0x3F0) >> 4);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* SZE */
   case 2: /* SOO */
   case 4: /* SOD */
   case 6: /* SMI */
   case 8: /* SNZ */
   case 10: /* SNO */
   case 12: /* SEV */
   case 14: /* SPL */
      sprintf (outbuf, "\t%s\t%s", ops4op[op], regnames[inst & 0x0007]);
      return (0);

   case 1: /* SSE */
   case 9: /* SSN */
   case 32: /* IDL */
      sprintf (outbuf, "\t%s\t%d", ops4op[op], inst & 0x000F);
      return (0);

   case 16: /* SLT */
   case 18: /* SEQ */
   case 20: /* SGT */
   case 22: /* SOV */
   case 24: /* SGE */
   case 26: /* SNE */
   case 28: /* SLE */
   case 30: /* SNV */
      break;

   case 54: /* SOC */
      op = 33;
      break;

   case 62: /* SNC */
      op = 34;
      break;

   default:
      return (-1);
   }

   sprintf (outbuf, "\t%s", ops4op[op]);
   return (0);
}

/***********************************************************************
* proc5op - Process type 5 operations.
***********************************************************************/

static int
proc5op (uint16 inst)
{
   Memory *operctl;
   SymNode *sym;
   uint16 oper;
   uint16 dev;
   uint16 op;
   char operand[32];

   dev = (inst & 0x065F);
   op = ((inst & 0x00F0) >> 4);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   if ((inst & 0xD9A0) == 0xD800) /* RDS */
      op = 0;
   else if ((inst & 0xD9A0) == 0xD820) /* WDS */
      op = 2;
   else switch (op)
   {
   case 8: /* LSB */
   case 9: /* LSR */
   case 10: /* LRF */
   case 12: /* SSB */
   case 14: /* SRF */
      op = op;
      break;

   default:
      return (-1);
   }

   operctl = getctl();
   if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
      return (-1);
   extended = TRUE;
   extendaddr = pcreg;
   oper = getinst();
   procmemop (operand, operctl, oper);

   if (op < 4)
      sprintf (outbuf, "\t%s\t>%04X\n\tDATA\t%s",
	       ops5op[op], dev, operand);
   else
      sprintf (outbuf, "\t@%s\t%s",
	       ops5op[op], operand);
   return (0);
}

/***********************************************************************
* proc6op - Process type 6 operations.
***********************************************************************/

static int
proc6op (uint16 inst)
{
   Memory *operctl;
   SymNode *sym;
   uint16 oper;
   uint16 op;
   uint16 bit;
   char operand[32];

   bit = inst & 0x000F;
   op = ((inst & 0x00F0) >> 4);

#ifdef DEBUGOPS1
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* TABZ */
   case 1: /* TABO */
   case 4: /* SABZ */
   case 5: /* SABO */
      sprintf (outbuf, "\t%s\t%d", ops6op[op], bit);
      break;

   case 2: /* TMBZ */
   case 3: /* TMBO */
   case 6: /* SMBZ */
   case 7: /* SMBO */
      operctl = getctl();
      if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
	 return (-1);
      extended = TRUE;
      extendaddr = pcreg;
      oper = getinst();
      procmemop (operand, operctl, oper);
      sprintf (outbuf, "\t%s\t%d,%s",
	       ops6op[op], bit, operand);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc7op - Process type 7 operations.
***********************************************************************/

static int
proc7op (uint16 inst)
{
   Memory *operctl;
   SymNode *sym;
   uint16 oper;
   uint16 op = 0;
   uint16 dev;
   char operand[32];

   dev = inst & 0x00FF;

   if ((inst & 0xFF00) == 0xD900) /* ATI */
      op = 0;
   else if ((inst & 0xFF00) == 0xDD00) /* API */
      op = 1;
   else return (-1);

   operctl = getctl();
   if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
      return (-1);
   extended = TRUE;
   extendaddr = pcreg;
   oper = getinst();
   procmemop (operand, operctl, oper);
   sprintf (outbuf, "\t%s\t>%02X\n\tDATA\t%s",
            ops7op[op], dev, operand);
   return (0);
}

/***********************************************************************
* proc8op - Process type 8 operations.
***********************************************************************/

static int
proc8op (uint16 inst)
{
   uint16 op = 0;

   if ((inst & 0xFF80) == 0xDF00) /* MVC */
      op = 0;
   else if ((inst & 0xFF80) == 0xDF80) /* CLC */
      op = 1;
   else return (-1);

   sprintf (outbuf, "\t%s", ops8op[op]);
   return (0);
}

/***********************************************************************
* dis9op - Process type 9 operations.
***********************************************************************/

static int
proc9op (uint16 inst)
{
   uint16 op = 0;
   uint16 sr, dr;
   uint16 smode, dmode;

   sr = ((inst >> 4) & 0xF);
   smode = (inst >> 10) & 0x3;
   if (smode < 3) sr |= 0x8;
   dr = (inst & 0xF);
   dmode = (inst >> 8) & 0x3;
   if (dmode < 3) dr |= 0x8;

   sprintf (outbuf, "\t%s\t%s%s,%s%s",
   	    "MOV", movmodes[smode], regnames[sr],
	    movmodes[dmode], regnames[dr]);
   return (0);
}

/***********************************************************************
* dis10op - Process type 10 operations.
***********************************************************************/

static int
proc10op (uint16 inst)
{
   Memory *operctl;
   uint16 oper;
   uint16 sr, dr;
   char operand[32];

   sr = ((inst >> 4) & 7);
   dr = (inst & 7);

   if ((inst & 0x0088) == 0x0088) /* LAR */
   {
      operctl = getctl();
      if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
	 return (-1);
      extended = TRUE;
      extendaddr = pcreg;
      oper = getinst();
      procmemop (operand, operctl, oper);
      sprintf (outbuf, "\t@%s\t%s", ops10op[0], operand);
   }
   else if ((inst & 0x0088) == 0x0080) /* SAR */
   {
      operctl = getctl();
      if (operctl->tag != ABSDATA_TAG && operctl->tag != RELDATA_TAG)
	 return (-1);
      extended = TRUE;
      extendaddr = pcreg;
      oper = getinst();
      procmemop (operand, operctl, oper);
      sprintf (outbuf, "\t@%s\t%s", ops10op[1], operand);
   }
   else if ((inst & 0x0088) == 0x0008) /* RFI */
   {
      sprintf (outbuf, "\t%s\t%s,%s",
	       ops10op[2], regnames[8+sr], regnames[8+dr]);
   }
   else if ((inst & 0x0088) == 0x0000) /* RFD */
   {
      sprintf (outbuf, "\t%s\t%s,%s",
	       ops10op[3], regnames[8+sr], regnames[8+dr]);
   }
   else
      return (-1);
   
   return (0);
}

/***********************************************************************
* dis11op - Process type 11 operations.
***********************************************************************/

static int
proc11op (uint16 inst)
{
   uint16 op = 0;

   op = (inst & 0x0F00) >> 8;

   switch (op)
   {
   case 0:
   case 1:
   case 2:
      sprintf (outbuf, "\t%s\t>%02X", ops11op[op], inst & 0xFF);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* procinst - process an instruction.
***********************************************************************/

static void 
procinst (uint16 inst, uint16 curpc)
{
   int type;
   int error;
   uint8 ic1, ic2;

   error = -1;
   extended = FALSE;
   type = decodeinst (inst);

   if (type != TYPE_ILL)
      error = inst_proc[type].proc (inst);

   if (error < 0)
      sprintf (outbuf, "\tDATA\t>%X", inst);

   printf ("%s", outbuf);
#if 1
   ic1 = (inst >> 8) & 0xFF;
   ic2 = inst & 0xFF;
   if (isprint (ic1) && !islower(ic1) && isprint(ic2) && !islower(ic2))
   {
      printf ("\t\tDATA\t'");
      if (ic1 == '\'')
         printf ("''");
      else
         printf ("%c", ic1);
      if (ic2 == '\'')
         printf ("'''");
      else
         printf ("%c'", ic2);
   }
#endif
   printf ("\n");

   if (extended)
   {
      printvals (extendaddr, -1);
   }
}

/***********************************************************************
* disprog - Disassemble a program.
***********************************************************************/

void
disprog (int junk)
{
   SymNode *s;
   uint16 inst;
   uint16 curpc;
   char lclidt[MAXSYMLEN+2];

#ifdef DEBUGOPS
   fprintf (stderr, "disprog: pcreg = %d, pgmlen = %d\n",
	    pcreg, pgmlen);
#endif

   strcpy (lclidt, idtbuf);
   if (strlen (lclidt) > 3) lclidt[3] = '\0';

   curpc = 0;
   if (memctl[curpc].tag == 0)
   {
      while (memctl[curpc].tag == 0) curpc++;
      printf ("\tORG\t>%X\n", curpc);
      pcreg = curpc;
   }

   while (pcreg < pgmlen)
   {
      SymNode *sym;
      Memory *instctl;

      curpc = pcreg;

      printvals (pcreg, 0);

      instctl = getctl();
      inst = getinst();

      if (instctl->tag == ABSDATA_TAG)
      {
#ifdef DEBUGOPS
	 fprintf (stderr, "%04X  %04X   ABSDATA\n", curpc, inst);
#endif
	 procinst (inst, curpc);
      }
      else if (instctl->tag == RELDATA_TAG)
      {

#ifdef DEBUGOPS
	 fprintf (stderr, "%04X  %04X   RELDATA\n", curpc, inst);
#endif
	 if (instctl->external)
	    sym = extlookup (inst, cardmode);
	 else 
	    sym = vallookup (inst, TRUE, FALSE);

	 if (sym)
	    printf ("\tDATA\t%s\n", sym->symbol);
	 else
	    printf ("\tDATA\t>%04X\n", inst);
      }
      else if (instctl->tag == RELORG_TAG)
      {
	 uint16 bsssize;
	 uint16 bsspc;

#ifdef DEBUGOPS
	 fprintf (stderr, "%04X  %04X   RELORG\n", curpc, inst);
#endif
	 bsssize = inst - curpc;
	 for (bsspc = pcreg; bsspc < inst; bsspc++)
	 {
	    if ((sym = vallookup (bsspc, FALSE, FALSE)) != NULL)
	    {
	       printf ("%s\tEQU\t$+%d\n", sym->symbol, bsspc-curpc);
	    }
	 }
	 printf ("\tBSS\t%d\n", bsssize);
	 pcreg = inst;
      }
      else if (instctl->tag == ABSORG_TAG)
      {
	 uint16 bsssize;
	 uint16 bsspc;

#ifdef DEBUGOPS
	 fprintf (stderr, "%04X  %04X   ABSORG\n", curpc, inst);
#endif
	 bsssize = inst - curpc;
	 for (bsspc = pcreg; bsspc < inst; bsspc++)
	 {
	    if ((sym = vallookup (bsspc, FALSE, FALSE)) != NULL)
	    {
	       printf ("%s\tEQU\t$+%d\n", sym->symbol, bsspc-curpc);
	    }
	 }
	 printf ("\tBSS\t%d\n", bsssize);
	 pcreg = inst;
      }
   }
   printvals (pcreg, 0);
}
