/***********************************************************************
*
* SEDIT - Forth screen editor.
*
* Changes:
*      08/25/11   DGP   Original.
*
***********************************************************************/
 
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#define NORMAL		0
#define ABORT		16

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#include "simdsk.h"

#define SECPSCR 8		/* Sectors per screen */
#define MAXSECTLEN 256		/* Max length of a sector */
#define LINELEN 64		/* Screen line length */

#define EDITOR "/bin/vi"

typedef struct
{
   int cyls;
   int heads;
   int secttrk;
   int sectlen;
   int overhead;
   int tracklen;
   int curcyl;
   int curhead;
   int cursect;
} DiskInfo;

static DiskInfo diskinfo;

/***********************************************************************
* dskreadint - Read an integer.
***********************************************************************/

static int
dskreadint (FILE *fd)
{
   int r;
   int i;

   r = 0;
   for (i = 0; i < 4; i++)
   {
      int c;
      if ((c = fgetc (fd)) < 0)
         return (-1);
      r = (r << 8) | (c & 0xFF);
   }
   return (r);
}

/***********************************************************************
* dskopen - Open disk and read geometry.
***********************************************************************/

FILE *
dskopen (char *dskname)
{
   FILE *fd;
   int diskdata;

#ifdef DEBUGDSK
   fprintf (stderr, "dskopen: dskname = %s\n", dskname);
#endif

   if ((fd = fopen (dskname, "r+b")) == NULL)
   {
      fprintf (stderr, "dskopen: Can't open disk: %s\n", dskname);
      return (NULL);
   }

   /*
   ** Seek to beginning of the disk
   */

   if (fseek (fd, 0, SEEK_SET) < 0)
   {
      fprintf (stderr, "dskopen: seek failed: %s", strerror (errno));
      return (NULL);
   }

   /*
   ** Read the disk geometry
   */

   if ((diskinfo.cyls = dskreadint (fd)) < 0) return (NULL);
   if ((diskinfo.heads = dskreadint (fd)) < 0) return (NULL);
   if ((diskdata = dskreadint (fd)) < 0) return (NULL);
   diskinfo.secttrk = diskdata & 0xFF;
   diskinfo.overhead = 0;
   if ((diskinfo.sectlen = dskreadint (fd)) < 0) return (NULL);
   diskinfo.tracklen = diskinfo.secttrk * 
	       (diskinfo.sectlen + diskinfo.overhead);
	       

#ifdef DEBUGDSK
   fprintf (stderr, "Disk geometry:\n");
   fprintf (stderr, "   cyls      = %d\n", diskinfo.cyls);
   fprintf (stderr, "   heads     = %d\n", diskinfo.heads);
   fprintf (stderr, "   secttrk   = %d\n", diskinfo.secttrk);
   fprintf (stderr, "   sectlen   = %d\n", diskinfo.sectlen);
   fprintf (stderr, "   overhead  = %d\n", diskinfo.overhead);
   fprintf (stderr, "   tracklen  = %d\n", diskinfo.tracklen);
   fprintf (stderr, "   disk size = %d bytes\n", diskinfo.tracklen *
	       diskinfo.heads * diskinfo.cyls);
#endif

   return (fd);
}

/***********************************************************************
* dskseek - Seek to specified disk location.
***********************************************************************/

static int
dskseek (FILE *fd, int cyl, int head, int sector)
{
   int dskloc;

   /*
   ** Ensure cylinder is OK
   */

   if ((cyl >= diskinfo.cyls) || (head >= diskinfo.heads) ||
       (sector >= diskinfo.secttrk))
   {
#ifdef DEBUGDSK
      fprintf (stderr, "dskseek: ERROR: cyl = %d, head = %d, sector = %d\n",
	       cyl, head, sector);
#endif
      return (-1);
   }

   /*
   ** Calculate location
   */

   dskloc = (diskinfo.tracklen * head) +
      (diskinfo.tracklen * diskinfo.heads * cyl) +
      ((diskinfo.overhead + diskinfo.sectlen) * sector) +
      DSKOVERHEAD;

#ifdef DEBUGDSK
   fprintf (stderr, "dskseek: cyl = %d, head = %d, sector = %d\n",
	    cyl, head, sector);
   fprintf (stderr, "   dskloc = %d\n", dskloc);
#endif

   /*
   ** Go to the specified location on disk
   */

   if (fseek (fd, dskloc, SEEK_SET) < 0)
   {
      fprintf (stderr, "dskseek: seek failed: %s", strerror (errno));
      return (-1);
   }

   return (0);
}

/***********************************************************************
* Main procedure
***********************************************************************/

int
main (int argc, char **argv)
{
   FILE *fd, *tfd;
   int i;
   int cyl, sect, scrnum;
   int firstsec;
   char tmpname[256];
   char tmpbuffer[MAXSECTLEN];
   char screenbuf[SECPSCR][MAXSECTLEN];

   if (argc != 3)
   {
      fprintf (stderr, "usage: sedit forthdisk screen\n");
      exit (1);
   }

   if ((fd = dskopen (argv[1])) == NULL)
      exit (1);

   /*
   ** Convert screen number to disk location.
   */

   scrnum = atoi (argv[2]);
   firstsec = scrnum * SECPSCR;
   cyl = firstsec / diskinfo.secttrk;
   sect = firstsec % diskinfo.secttrk;

   /*
   ** Read in the screen
   */

   for (i = 0; i < SECPSCR; i++)
   {
      if (dskseek (fd, cyl, 0, sect) < 0)
      {
	 fprintf (stderr, "sedit: disk seek error\n");
	 fclose (fd);
	 exit (1);
      }

      if (fread (tmpbuffer, 1, diskinfo.sectlen, fd) != diskinfo.sectlen)
      {
	 fprintf (stderr, "sedit: disk read error\n");
	 fclose (fd);
	 exit (1);
      }

      /* Each sector has two "lines" */
      memcpy (screenbuf[i], tmpbuffer, LINELEN);
      memcpy (&screenbuf[i][LINELEN+1], &tmpbuffer[LINELEN], LINELEN);
      screenbuf[i][LINELEN] = '\n';
      screenbuf[i][LINELEN+LINELEN+1] = '\n';
      screenbuf[i][LINELEN+LINELEN+2] = 0;

      sect++;
      if (sect >= diskinfo.secttrk)
      {
         cyl++;
	 sect = 0;
      }
   }
#ifdef DEBUGDSK
   printf ("SCR # %d\n", scrnum);
   for (i = 0; i < SECPSCR; i ++)
      printf ("%s", screenbuf[i]);
   exit (0);
#endif

   /*
   ** Write to tmp file and launch editor.
   */

   sprintf (tmpname, "ED%08d.tmp", getpid());
   if ((tfd = fopen (tmpname, "w")) == NULL)
   {
      fprintf (stderr, "sedit: Can't open temp file: %s\n", tmpname);
      fclose (fd);
      exit (1);
   }
   for (i = 0; i < SECPSCR; i ++)
      fprintf (tfd, "%s", screenbuf[i]);
   fclose (tfd);

   sprintf (tmpbuffer, "%s %s", EDITOR, tmpname);
   system (tmpbuffer);

   /*
   ** Read edited file back into screen buffer.
   */

   if ((tfd = fopen (tmpname, "r")) == NULL)
   {
      fprintf (stderr, "sedit: Can't open temp file: %s\n", tmpname);
      fclose (fd);
      exit (1);
   }

   for (i = 0; i < SECPSCR; i ++)
   {
      char *bp;

      memset (tmpbuffer, ' ', sizeof(tmpbuffer));
      fgets (tmpbuffer, sizeof(tmpbuffer), tfd);
      bp = strchr (tmpbuffer, '\n');
      while ((bp-tmpbuffer) < LINELEN) *bp++ = ' ';
      strncpy (screenbuf[i], tmpbuffer, LINELEN);

      memset (tmpbuffer, ' ', sizeof(tmpbuffer));
      fgets (tmpbuffer, sizeof(tmpbuffer), tfd);
      bp = strchr (tmpbuffer, '\n');
      while ((bp-tmpbuffer) < LINELEN) *bp++ = ' ';
      strncpy (&screenbuf[i][LINELEN], tmpbuffer, LINELEN);
   }
   fclose (tfd);

   /*
   ** Write screen buffer back to disk.
   */

   firstsec = scrnum * SECPSCR;
   cyl = firstsec / diskinfo.secttrk;
   sect = firstsec % diskinfo.secttrk;

   for (i = 0; i < SECPSCR; i++)
   {
      if (dskseek (fd, cyl, 0, sect) < 0)
      {
	 fprintf (stderr, "sedit: disk seek error\n");
	 fclose (fd);
	 exit (1);
      }

      if (fwrite (screenbuf[i], 1, diskinfo.sectlen, fd) != diskinfo.sectlen)
      {
	 fprintf (stderr, "sedit: disk write error\n");
	 fclose (fd);
	 exit (1);
      }

      sect++;
      if (sect >= diskinfo.secttrk)
      {
         cyl++;
	 sect = 0;
      }
   }

   unlink (tmpname);
   fclose (fd);
}
